import asyncio
import websockets
import logging
import ssl
import json
import time
import datetime
from datetime import timezone

from indralib.indra_event import IndraEvent
from indralib.indra_time import IndraTime
from indralib.server_config import Profiles


class IndraClient:
    def __init__(
        self,
        profile=None,
        profile_name="default",
        verbose=False,
        module_name=None,
        log_handler=None,
    ):
        self.log = logging.getLogger("IndraClient")
        if log_handler is not None:
            self.log.addHandler(log_handler)
        if module_name is None:
            self.module_name = "IndraClient (python)"
        else:
            self.module_name = module_name
        self.websocket = None
        self.verbose = verbose
        self.trx = {}
        self.recv_queue = asyncio.Queue()
        self.recv_task = None
        self.initialized = False
        self.error_shown = False
        self.profiles = Profiles()
        if profile is not None and Profiles.check_profile(profile) is False:
            self.log.error(f"Invalid profile {profile}")
            self.profile = None
            return
        if profile is not None:
            self.profile = profile
            self.initialized = True
        else:
            if profile_name == "default":
                self.profile = self.profiles.get_default_profile()
            else:
                self.profile = self.profiles.get_profile(profile_name)
            if self.profile is not None:
                self.initialized = True

    async def init_connection(self, verbose=False):
        """Initialize connection"""
        if self.initialized is False:
            self.trx = {}
            self.websocket = None
            self.log.error(
                "Indrajala init_connection(): connection profile data not initialized!"
            )
            return None
        if self.websocket is not None:
            if verbose is True:
                self.log.warning(
                    "Websocket already initialized, please call close_connection() first!"
                )
            return self.websocket
        self.trx = {}
        self.uri = Profiles.get_uri(self.profile)
        if self.profile["TLS"] is True:
            ssl_ctx = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
            if self.profile["ca_authority"] is not None:
                try:
                    ssl_ctx.load_verify_locations(cafile=self.profile["ca_authority"])
                except Exception as e:
                    self.log.error(
                        f"Could not load CA authority file {self.profile['ca_authority']}: {e}"
                    )
                    self.websocket = None
                    return None
        try:
            self.websocket = await websockets.connect(self.uri, ssl=ssl_ctx)
        except Exception as e:
            self.log.error(f"Could not connect to {self.uri}: {e}")
            self.websocket = None
            return None
        self.recv_queue.empty()
        self.recv_task = asyncio.create_task(self.fn_recv_task())
        self.session_id = None
        self.username = None
        self.error_shown = False
        return self.websocket

    async def fn_recv_task(self):
        """Receive task"""
        while True:
            try:
                message = await self.websocket.recv()
                if self.verbose is True:
                    self.log.info(f"Received message: {message}")
            except Exception as e:
                self.log.error(f"Could not receive message: {e}, exiting recv_task()")
                self.recv_task = None
                return False
            # ie = IndraEvent()
            ie = IndraEvent.from_json(message)
            if ie.uuid4 in self.trx:
                fRec = self.trx[ie.uuid4]
                dt = time.time() - fRec["start_time"]
                if self.verbose is True:
                    self.log.info(
                        "---------------------------------------------------------------"
                    )
                    self.log.info(
                        f"Future: trx event {ie.to_scope}, uuid: {ie.uuid4}, {ie.data_type}, dt={dt}"
                    )
                fRec["future"].set_result(ie)
                del self.trx[ie.uuid4]
            else:
                if self.verbose is True:
                    self.log.info(
                        f"Received event {ie.to_scope}, uuid: {ie.uuid4}, {ie.data_type}"
                    )
                await self.recv_queue.put(ie)
        self.recv_task = None
        return

    async def send_event(self, event):
        """Send event"""
        if self.initialized is False:
            if self.error_shown is False:
                self.error_shown = True
                self.log.error(
                    "Indrajala send_event(): connection data not initialized!"
                )
            return False
        if self.websocket is None:
            self.log.error(
                "Websocket not initialized, please call init_connection() first!"
            )
            return False
        if isinstance(event, IndraEvent) is False:
            self.log.error("Please provide an IndraEvent object!")
            return False
        if event.domain.startswith("$trx/") is True:
            replyEventFuture = asyncio.futures.Future()
            fRec = {
                "future": replyEventFuture,
                "start_time": time.time(),
            }
            self.trx[event.uuid4] = fRec
            self.log.debug("Future: ", replyEventFuture)
        else:
            replyEventFuture = None
        try:
            await self.websocket.send(event.to_json())
        except Exception as e:
            self.log.error(f"Could not send message: {e}")
            self.initialized = False
        return replyEventFuture

    async def recv_event(self, timeout=None):
        """Receive event"""
        if self.initialized is False:
            self.log.error("Indrajala recv_event(): connection data not initialized!")
            return None
        if self.websocket is None:
            self.log.error(
                "Websocket not initialized, please call init_connection() first!"
            )
            return None
        if timeout is None:
            try:
                ie = await self.recv_queue.get()
            except Exception as e:
                self.log.error(f"Could not receive message: {e}")
                return None
        else:
            try:
                ie = await asyncio.wait_for(self.recv_queue.get(), timeout=timeout)
            except TimeoutError:
                return None
            except Exception as e:
                self.log.warning(f"Timeout receive failed: {e}")
                return None
        self.recv_queue.task_done()
        return ie

    async def close_connection(self):
        """Close connection"""
        if self.initialized is False:
            self.log.error(
                "Indrajala close_connection(): connection data not initialized!"
            )
            return False
        if self.websocket is None:
            self.log.error(
                "Websocket not initialized, please call init_connection() first!"
            )
            return False
        if self.recv_task is not None:
            self.recv_task.cancel()
            self.recv_task = None
        await self.websocket.close()
        self.trx = {}
        self.websocket = None
        self.session_id = None
        self.username = None
        return True

    async def subscribe(self, domains):
        """Subscribe to domain"""
        if self.initialized is False:
            self.log.error("Indrajala subscribe(): connection data not initialized!")
            return False
        if self.websocket is None:
            self.log.error(
                "Websocket not initialized, please call init_connection() first!"
            )
            return False
        if domains is None or domains == "":
            self.log.error("Please provide a valid domain(s)!")
            return False
        ie = IndraEvent()
        ie.domain = "$cmd/subs"
        ie.from_id = "ws/python"
        ie.data_type = "vector/string"
        ie.auth_hash = self.session_id
        if isinstance(domains, list) is True:
            ie.data = json.dumps(domains)
        else:
            ie.data = json.dumps([domains])
        await self.websocket.send(ie.to_json())
        return True

    async def unsubscribe(self, domains):
        """Unsubscribe from domain"""
        if self.initialized is False:
            self.log.error("Indrajala unsubscribe(): connection data not initialized!")
            return False
        if self.websocket is None:
            self.log.error(
                "Websocket not initialized, please call init_connection() first!"
            )
            return False
        if domains is None or domains == "":
            self.log.error("Please provide a valid domain(s)!")
            return False
        ie = IndraEvent()
        ie.domain = "$cmd/unsubs"
        ie.from_id = "ws/python"
        ie.data_type = "vector/string"
        ie.auth_hash = self.session_id
        if isinstance(domains, list) is True:
            ie.data = json.dumps(domains)
        else:
            ie.data = json.dumps([domains])
        await self.websocket.send(ie.to_json())
        return True

    async def get_history(
        self, domain, start_time=None, end_time=None, sample_size=None, mode="Sample"
    ):
        """Get history of domain

        returns a future object, which will be set when the reply is received
        """
        cmd = {
            "domain": domain,
            "time_jd_start": start_time,
            "time_jd_end": end_time,
            "limit": sample_size,
            # "data_type": "number/float%",
            "mode": mode,
        }
        ie = IndraEvent()
        ie.domain = "$trx/db/req/history"
        ie.from_id = "ws/python"
        ie.data_type = "historyrequest"
        ie.auth_hash = self.session_id
        ie.data = json.dumps(cmd)
        if self.verbose is True:
            self.log.info(f"Sending: {ie.to_json()}")
        return await self.send_event(ie)

    async def get_wait_history(
        self, domain, start_time=None, end_time=None, sample_size=None, mode="Sample"
    ):
        future = await self.get_history(domain, start_time, end_time, sample_size, mode)
        hist_result = await future
        return json.loads(hist_result.data)

    @staticmethod
    def get_current_time_jd():
        cdt = datetime.datetime.now(timezone.utc)
        dt_jd = IndraTime.datetime2julian(cdt)
        return dt_jd

    @staticmethod
    async def _get_history_block(
        username=None,
        password=None,
        domain=None,
        start_time=None,
        end_time=None,
        verbose=False,
        sample_size=20,
    ):
        ic = IndraClient(verbose=verbose)
        if verbose is True:
            ic.log.info(f"Connecting")
        await ic.init_connection()
        if verbose is True:
            ic.log.info("Connected to Indrajala")
        ret = await ic.login_wait(username, password)
        if verbose is True:
            ic.log.info(f"Login result: {ret}")
        if ret is not None:
            if verbose is True:
                ic.log.info(f"Logged in as {username}")
            data = await ic.get_wait_history(
                domain=domain,
                start_time=start_time,
                end_time=end_time,
                sample_size=sample_size,
                mode="Sample",
            )
            if verbose is True:
                ic.log.info(f"Data: {data}")
            await ic.logout_wait()
            await ic.close_connection()
            if verbose is True:
                ic.log.info("Logged out and closed connection")
            return data
        return None

    @staticmethod
    def get_history_sync(
        username,
        password,
        domain,
        start_time,
        end_time=None,
        sample_size=20,
        verbose=False,
    ):
        return asyncio.run(
            IndraClient._get_history_block(
                username, password, domain, start_time, end_time, verbose, sample_size
            )
        )

    async def get_last_event(self, domain):
        """Get last event of domain"""
        ie = IndraEvent()
        ie.domain = "$trx/db/req/last"
        ie.from_id = "ws/python"
        ie.data_type = "json/reqlast"
        ie.auth_hash = self.session_id
        ie.data = json.dumps({"domain": domain})
        return await self.send_event(ie)

    async def get_wait_last_event(self, domain):
        future = await self.get_last_event(domain)
        last_result = await future
        if last_result.data is not None and last_result.data != "":
            return IndraEvent.from_json(last_result.data)
        else:
            return None

    async def get_unique_domains(self, domain=None, data_type=None):
        """Get unique domains"""
        if domain is None:
            domain = "$event/%"
        if data_type is None:
            data_type = "%"
        cmd = {
            "domain": domain,
            "data_type": data_type,
        }
        ie = IndraEvent()
        ie.domain = "$trx/db/req/uniquedomains"
        ie.from_id = "ws/python"
        ie.data_type = "uniquedomainsrequest"
        ie.auth_hash = self.session_id
        ie.data = json.dumps(cmd)
        return await self.send_event(ie)

    async def get_wait_unique_domains(self, domain=None, data_type=None):
        future = await self.get_unique_domains(domain, data_type)
        domain_result = await future
        return json.loads(domain_result.data)

    async def delete_recs(self, domains=None, uuid4s=None):
        if domains is None and uuid4s is None:
            self.log.error("Please provide a domain or uuid4s")
            return False
        if domains is not None and uuid4s is not None:
            self.log.error("Please provide either a domain or uuid4s")
            return False
        cmd = {
            "domains": domains,
            "uuid4s": uuid4s,
        }
        ie = IndraEvent()
        ie.domain = "$trx/db/req/del"
        ie.from_id = "ws/python"
        ie.data_type = "json/reqdel"
        ie.auth_hash = self.session_id
        ie.data = json.dumps(cmd)
        return await self.send_event(ie)

    async def delete_recs_wait(self, domains=None, uuid4s=None):
        future = await self.delete_recs(domains, uuid4s)
        result = await future
        if result.data_type.startswith("error") is True:
            self.log.error(f"Error: {result.data}")
            return None
        else:
            return json.loads(result.data)

    async def update_recs(self, recs):
        if isinstance(recs, list) is False:
            self.log.error("Not a list")
            recs = [recs]
        cmd = recs
        ie = IndraEvent()
        ie.domain = "$trx/db/req/update"
        ie.from_id = "ws/python"
        ie.data_type = "json/requpdate"
        ie.auth_hash = self.session_id
        ie.data = json.dumps(cmd)
        return await self.send_event(ie)

    async def update_recs_wait(self, recs):
        future = await self.update_recs(recs)
        result = await future
        if result.data_type.startswith("error") is True:
            self.log.error(f"Error: {result.data}")
            return None
        else:
            return json.loads(result.data)

    async def kv_write(self, key, value):
        cmd = {
            "key": key,
            "value": value,
        }
        ie = IndraEvent()
        ie.domain = "$trx/kv/req/write"
        ie.from_id = "ws/python"
        ie.data_type = "kvwrite"
        ie.auth_hash = self.session_id
        ie.data = json.dumps(cmd)
        return await self.send_event(ie)

    async def kv_write_wait(self, key, value):
        future = await self.kv_write(key, value)
        result = await future
        if result.data_type.startswith("error") is True:
            self.log.error(f"Error: {result.data}")
            return None
        else:
            return json.loads(result.data)

    async def kv_read(self, key):
        cmd = {
            "key": key,
        }
        ie = IndraEvent()
        ie.domain = "$trx/kv/req/read"
        ie.from_id = "ws/python"
        ie.data_type = "kvread"
        ie.auth_hash = self.session_id
        ie.data = json.dumps(cmd)
        self.log.debug("Sending kv_read")
        return await self.send_event(ie)

    async def kv_read_wait(self, key):
        future = await self.kv_read(key)
        result = await future
        if result.data_type.startswith("error") is True:
            self.log.error(f"Error: {result.data}")
            return None
        else:
            return json.loads(result.data)

    async def kv_delete(self, key):
        cmd = {
            "key": key,
        }
        ie = IndraEvent()
        ie.domain = "$trx/kv/req/delete"
        ie.from_id = "ws/python"
        ie.data_type = "kvdelete"
        ie.auth_hash = self.session_id
        ie.data = json.dumps(cmd)
        return await self.send_event(ie)

    async def kv_delete_wait(self, key):
        future = await self.kv_delete(key)
        result = await future
        if result.data_type.startswith("error") is True:
            self.log.error(f"Error: {result.data}")
            return None
        else:
            return json.loads(result.data)

    async def login(self, username, password):
        cmd = {
            "key": f"entity/indrajala/user/{username}/password",
            "value": password,
        }
        self.username = username
        ie = IndraEvent()
        ie.domain = "$trx/kv/req/login"
        ie.from_id = "ws/python"
        ie.data_type = "kvverify"
        ie.auth_hash = self.session_id
        ie.data = json.dumps(cmd)
        return await self.send_event(ie)

    async def login_wait(self, username, password):
        future = await self.login(username, password)
        result = await future
        if result.data_type.startswith("error") is True:
            self.log.error(f"Error: {result.data}")
            self.session_id = None
            self.username = None
            return None
        else:
            self.log.debug(
                f"Login result: {result.data}, {result.data_type}, {result.auth_hash}"
            )
            self.session_id = result.auth_hash
            return result.auth_hash

    async def logout(self):
        ie = IndraEvent()
        ie.domain = "$trx/kv/req/logout"
        ie.from_id = "ws/python"
        ie.auth_hash = self.session_id
        ie.data_type = ""
        ie.data = ""
        return await self.send_event(ie)

    async def logout_wait(self):
        future = await self.logout()
        result = await future
        if result.data_type.startswith("error") is True:
            self.log.error(f"Error: {result.data}")
            return False
        else:
            return True

    async def indra_log(self, level, message, module_name=None):
        """Log message"""
        if module_name is None:
            module_name = self.module_name
        if level not in ["debug", "info", "warn", "error"]:
            self.log.error(f"Invalid log level: {level}, {message}")
            return False
        ie = IndraEvent()
        ie.domain = f"$log/{level}"
        ie.from_id = f"ws/python/{module_name}"
        ie.data_type = "log"
        ie.auth_hash = self.session_id
        ie.data = json.dumps(message)
        return await self.send_event(ie)

    async def debug(self, message, module_name=None):
        self.log.debug(f"Indra_log-Debug: {message}")
        return await self.indra_log("debug", message, module_name)

    async def info(self, message, module_name=None):
        self.log.info(f"Indra_log-Info: {message}")
        return await self.indra_log("info", message, module_name)

    async def warn(self, message, module_name=None):
        self.log.warn(f"Indra_log-Warn: {message}")
        return await self.indra_log("warn", message, module_name)

    async def error(self, message, module_name=None):
        self.log.error(f"Indra_log-Error: {message}")
        return await self.indra_log("error", message, module_name)

    @staticmethod
    def get_timeseries(result):
        dt = []
        y = []
        for t, yv in result:
            dt.append(IndraTime.julian2datetime(t))
            y.append(yv)
        return dt, y



import os
import logging
from hashlib import md5
import urllib.request as request
import cgi
import pickle
import re
import io
import pandas as pd
import json
import datetime
import requests
import tomllib as toml
import uuid

from indralib.indra_event import IndraEvent
import io

# Requires: pandas, lxml, openpyxl


class IndraDownloader:
    def __init__(self, cache_dir="download_cache", use_cache=True):
        self.cache_dir = cache_dir
        self.use_cache = use_cache
        self.log = logging.getLogger("Downloader")
        if use_cache is True:
            self.cache_info = {}
            if os.path.isdir(cache_dir) is False:
                try:
                    os.makedirs(cache_dir)
                    self.log.debug(f"Cache directory {cache_dir} created.")
                except Exception as e:
                    self.use_cache = False
                    self.log.error(f"Failed to create cache {cache_dir}: {e}")
                    return
            self.cache_info_file = os.path.join(cache_dir, "cache_info.json")
            if os.path.exists(self.cache_info_file):
                try:
                    with open(self.cache_info_file, "r") as f:
                        self.cache_info = json.load(f)
                except Exception as e:
                    self.log.error(f"Failed to read cache_info: {e}")
            # Check for cache consistency, delete inconsistent entries:
            entries = list(self.cache_info.keys())
            for entry in entries:
                valid = True
                for mand in ["cache_filename", "time"]:
                    if mand not in self.cache_info[entry]:
                        self.log.warning(
                            f"Cache-entry for {entry} inconsistent: no {mand} field, deleting entry."
                        )
                        del self.cache_info[entry]
                        valid = False
                        break
                if valid is False:
                    continue
                lpath = os.path.join(
                    self.cache_dir, self.cache_info[entry]["cache_filename"]
                )
                if os.path.exists(lpath) is False:
                    self.log.warning(
                        f"Local file {lpath} for cache entry {entry} does not exist, deleting cache entry."
                    )
                    del self.cache_info[entry]
                    continue

    def update_cache(self, url, cache_filename):
        if self.use_cache:
            for other_url in self.cache_info:
                if other_url == url:
                    continue
                if "cache_filename" in self.cache_info[other_url]:
                    if self.cache_info[other_url]["cache_filename"] == cache_filename:
                        self.log.error(
                            "FATAL cache name clash: {other_url} and {url} both want to cache a file named {cache_filename}"
                        )
                        self.log.error(
                            "Caching will not work for {url}, cache for {other_url} delete too."
                        )
                        self.log.error("-----Algorithm must be changed!------")
                        del self.cache_info[other_url]
                        return
            self.cache_info[url] = {}
            self.cache_info[url]["cache_filename"] = cache_filename
            self.cache_info[url]["time"] = datetime.datetime.now().isoformat()
            try:
                with open(self.cache_info_file, "w") as f:
                    json.dump(self.cache_info, f, indent=4)
                    # self.log.info(f"Saved cache_info to {self.cache_info_file}")
            except Exception as e:
                self.log.error(f"Failed to update cache_info: {e}")

    def decode(self, encoding_name):
        data = self.tf_data
        return data.decode(encoding_name)

    def unpickle(self):
        data = self.tf_data
        return pickle.loads(data)

    def extract_lines(self, start, stop=0):
        data = self.tf_data
        lines = data.split("\n")
        if stop == 0:
            stop = len(lines)
        if start < 0:
            start = len(lines) + start
        if stop < 0:
            stop = len(lines) + stop
        if start < 1 or stop <= start:
            self.log.error(
                f"Format required: extract_lines(start_line_no:end_line_no), e.g. extract_lines(10:100)"
            )
            self.log.error(f"start_line_no >=1 and end_line_no>start_line_no")
            return None
        if stop > len(lines):
            self.log.error(
                f"Format required: extract_lines(start_line_no:end_line_no), e.g. extract_lines(10:100)"
            )
            self.log.error(
                f"end_line_no {stop} is > line-count in source file: {len(lines)}"
            )
            return None
        data = "\n".join(lines[start - 1 : stop])
        lno = len(data.split("\n"))
        self.log.debug(f"Extracted {lno} lines, [{start}:{stop}]")
        return data

    def remove_comments(self, comment_line_start="#"):
        data = self.tf_data
        lines = data.split("\n")
        valid_data = []
        for line in lines:
            if line.startswith(comment_line_start) is False:
                valid_data.append(line)
        valid_data = "\n".join(valid_data)
        return valid_data

    def extract_html_table(self, index):
        data = self.tf_data
        tables = pd.read_html(data)
        if len(tables) > index:
            return tables[index]
        else:
            lno = len(tables)
            self.log.error(f"No table with index {index}, table count is {lno}")
            return None

    def pandas_csv_separator(self, sep):
        data = self.tf_data
        if sep == " ":
            return pd.read_csv(io.StringIO(data), sep=r"\s+", engine="python")
        return pd.read_csv(io.StringIO(data), sep=sep, engine="python")

    def pandas_filter(self, column_list):
        data = self.tf_data
        return data.filter(column_list, axis=1)

    def pandas_csv_separator_nan(self, sep, nan):
        data = self.tf_data
        if sep == " ":
            return pd.read_csv(
                io.StringIO(data), sep=r"\s+", na_values=nan, engine="python"
            )
        return pd.read_csv(io.StringIO(data), sep=sep, na_values=nan, engine="python")

    def pandas_excel_rowskips(self, skiprow_list):
        data = self.tf_data
        # change to use BytesIO:
        data_bio = io.BytesIO(data)
        return pd.read_excel(data_bio, skiprows=skiprow_list)

    def pandas_excel_worksheet_subset(
        self, worksheet_name, include_rows, include_columns
    ):
        data = self.tf_data
        return pd.read_excel(
            data,
            sheet_name=worksheet_name,
            skiprows=lambda x: x + 1 not in range(include_rows[0], include_rows[1] + 1),
            usecols=include_columns,
        )

    def add_prefix(self, prefix):
        data = self.tf_data
        return prefix + "\n" + data

    def replace(self, token, replacement):
        data = self.tf_data
        return data.replace(token, replacement)

    def indra_import(self, time, column):
        self.log.info(f"Indra import: {time}, {column} NOT IMPLEMENTED YET")
        return self.tf_data

    def transform_df(self, df, transform, left, right):
        # XXX this is unsafe code.
        try:
            exec(transform)
        except Exception as e:
            self.log.error(f"Failed to apply transform {transform}: {e}")
            print(df.head(5))
            return None
        return df

    def single_transform(self, data, transform):
        if transform.startswith("df['"):
            ind = transform.find("=")
            if ind == -1:
                self.log.error(
                    f"Format required: df['column_name'] = transform, e.g. df['column_name'] = df['column_name'].str.lower()"
                )
                return None
            left = transform[:ind]
            right = transform[ind + 1 :]
            ind = left.find("']")
            if ind == -1:
                self.log.error(
                    f"Format required: df['column_name'] = transform, e.g. df['column_name'] = df['column_name'].str.lower()"
                )
                return None
            # check, if data is a dataframe
            if isinstance(data, pd.DataFrame) is False:
                self.log.error(f"Data is not a dataframe, cannot apply transform")
                return None
            return self.transform_df(data, transform, left, right)
        else:
            try:
                fn_name = transform.split("(")[0]
            except Exception as e:
                self.log.error(f"Failed to parse transform {transform}: {e}")
                return None
            tf = getattr(self, fn_name)
            if tf is not None:
                trs = "self." + transform
                self.tf_data = data
                try:
                    data = eval(trs)
                except Exception as e:
                    self.log.error(f"Failed to eval {trs}: {e}")
                    return None
            return data

    def transform(self, data, transforms):
        data_dict = {}
        if transforms is None:
            return data
        original_data = data
        for dataset_name in transforms:
            self.log.info(f"Creating dataset {dataset_name}")
            data = original_data
            for transform in transforms[dataset_name]:
                data = self.single_transform(data, transform)
            dataset = data
            if dataset is not None:
                data_dict[dataset_name] = dataset
        return data_dict

    def get(
        self,
        url,
        transforms=None,
        user_agent=None,
        resolve_redirects=True,
    ):
        cache_filename = None
        cache_path = None
        if resolve_redirects is True:
            try:
                # self.log.info(f"Test for redirect: {url}")
                r = requests.get(url, allow_redirects=True)
                # self.log.info(f"ReqInfo: {r}")
                if r.url != url:
                    self.log.warning(f"Redirect resolved: {url}->{r.url}")
                    url = r.url
            except Exception as e:
                self.log.info(f"Could not resolve redirects {e}")
        if self.use_cache is True:
            if url in self.cache_info:
                cache_filename = self.cache_info[url]["cache_filename"]
                cache_path = os.path.join(self.cache_dir, cache_filename)
                cache_time = self.cache_info[url]["time"]

        retrieved = False
        if cache_filename is None:
            try:
                remotefile = request.urlopen(url)
                remote_info = remotefile.info()
                # self.log.info(f"Remote.info: {remote_info}")
                if "Content-Disposition" in remote_info:
                    info = remote_info["Content-Disposition"]
                    value, params = cgi.parse_header(info)
                    # self.log.info(f"header: {params}")
                    if "filename" in params:
                        cache_filename = params["filename"]
                        cache_path = os.path.join(self.cache_dir, cache_filename)
                        self.log.info(f"Local filename is set to {cache_filename}")
                        self.log.info(f"Starting download via retrieve from {url}...")
                        request.urlretrieve(url, cache_path)
                        self.log.info(f"Download from {url}: OK.")
                        retrieved = True
            except Exception as e:
                cache_filename = None
            if retrieved is True:
                self.update_cache(url, cache_filename)
                return
        if cache_filename is None:
            url_comps = url.rsplit("/", 1)
            if len(url_comps) == 0:
                self.log.error(f"Invalid url {url}")
                return None
            fn = url_comps[-1]
            if "=" in fn:
                url_comps = fn.rsplit("=", 1)
            cache_filename = url_comps[-1]
            cache_path = os.path.join(self.cache_dir, cache_filename)
        if self.use_cache is True:
            if os.path.exists(cache_path):
                dl = False
                try:
                    with open(cache_path, "rb") as f:
                        data = f.read()
                        dl = True
                except Exception as e:
                    self.log.error(f"Failed to read cache {cache_path} for {url}: {e}")
                if dl is True:
                    self.log.info(f"Read {url} from cache at {cache_path}")
                    if len(data) > 0:
                        data = self.transform(data, transforms)
                        return data
                    else:
                        self.log.error(f"Ignoring zero-length cache-file {cache_path}")
                        dl = False
        self.log.info(f"Starting download from {url}...")
        data = None
        if user_agent is not None:
            req = request.Request(
                url, data=None, headers={"user-agent": user_agent, "accept": "*/*"}
            )
            self.log.info(f"Downloading with user_agent set to: {user_agent}")
            dl = False

            try:
                response = request.urlopen(req)
                data = response.read()
                dl = True
            except Exception as e:
                self.log.error(f"Failed to download from {url}: {e}")
                return None
        else:
            try:
                response = request.urlopen(url)
                data = response.read()
            except Exception as e:
                self.log.error(f"Failed to download from {url}: {e}")
                return None
        self.log.info(f"Download from {url}: OK.")
        if self.use_cache is True:
            try:
                with open(cache_path, "wb") as f:
                    f.write(data)
                    self.update_cache(url, cache_filename)
            except Exception as e:
                self.log.warning(
                    f"Failed to save to cache at {cache_path} for {url}: {e}"
                )
        data = self.transform(data, transforms)
        return data

    def get_datasets(self, data_sources_dir):
        dfs = {}
        self.indra_imports = {}
        for file in os.listdir(data_sources_dir):
            if file.endswith(".toml"):
                filepath = os.path.join(data_sources_dir, file)
                self.log.info(f"processing: {filepath}")
                try:
                    with open(filepath, "rb") as f:
                        data_desc = toml.load(f)
                except Exception as e:
                    self.log.error(f"Failed to read toml file {filepath}: {e}")
                    continue
                req = ["citation/data_source", "datasets"]
                for r in req:
                    pt = r.split("/")
                    if len(pt) == 1:
                        if pt[0] not in data_desc:
                            self.log.error(
                                f"{filepath} doesn't have [{pt[0]}] section."
                            )
                            continue
                        continue
                    if len(pt) != 2:
                        self.log.error(f"req-field doesn't parse: {r}")
                        continue
                    if pt[0] not in data_desc:
                        self.log.error(f"{filepath} doesn't have [{pt[0]}] section.")
                        continue
                    if pt[1] not in data_desc[pt[0]]:
                        self.log.error(
                            f"{filepath} doesn't have a {pt[1]}= entry in [{pt[0]}] section."
                        )
                        continue
                # print("----------------------------------------------------------------------------------")
                # print(f"Processing {filepath}")
                if "user_agent" in data_desc["citation"]:
                    ua = data_desc["citation"]["user_agent"]
                else:
                    ua = None
                if "redirect" in data_desc["citation"]:
                    use_redirect = data_desc["citation"]["redirect"]
                else:
                    use_redirect = True
                if "indra_import" in data_desc:
                    indra_imports = data_desc["indra_import"]
                    self.log.info(f"indra_imports: {indra_imports}")
                else:
                    indra_imports = None
                data_dicts = self.get(
                    data_desc["citation"]["data_source"],
                    transforms=data_desc["datasets"],
                    user_agent=ua,
                    resolve_redirects=use_redirect,
                )
                if data_dicts is None:
                    self.log.error(
                        f"Failed to retrieve dataset(s) from {data_desc['citation']['data_source']}"
                    )
                    continue
                for dataset in data_dicts:
                    # print(f">>> {dataset}")
                    data = data_dicts[dataset]
                    # if type(data)==str:
                    #     print(data)
                    # else:
                    #     print(data.head())
                    #     print("...")
                    #     print(data.tail())
                    dfs[dataset] = {}
                    dfs[dataset]["data"] = data
                    dfs[dataset]["metadata"] = data_desc["citation"]
        return dfs
